from typing import Any, Callable, Dict, Iterable

import werkzeug
from werkzeug import exceptions

import pjrpc.server


class JsonRPC:
    """
    `werkzeug <https://werkzeug.palletsprojects.com/en/0.16.x/>`_ server JSON-RPC integration.

    :param path: JSON-RPC handler base path
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
    """

    def __init__(self, path: str = '', **kwargs: Any):
        self._path = path
        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)

    def __call__(self, environ: Dict[str, Any], start_response: Callable[..., Any]) -> Iterable[bytes]:
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ: Dict[str, Any], start_response: Callable[..., Any]) -> Iterable[bytes]:
        environ['app'] = self
        request = werkzeug.Request(environ)
        response = self._rpc_handle(request)
        return response(environ, start_response)

    @property
    def dispatcher(self) -> pjrpc.server.Dispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    def _rpc_handle(self, request: werkzeug.Request) -> werkzeug.Response:
        """
        Handles JSON-RPC request.

        :returns: werkzeug response
        """

        if request.content_type not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise exceptions.UnsupportedMediaType()

        try:
            request_text = request.get_data(as_text=True)
        except UnicodeDecodeError as e:
            raise exceptions.BadRequest() from e

        response_text = self._dispatcher.dispatch(request_text, context=request)
        if response_text is None:
            return werkzeug.Response()
        else:
            return werkzeug.Response(response_text, mimetype=pjrpc.common.DEFAULT_CONTENT_TYPE)



import abc
import functools as ft
import json
import logging
from types import SimpleNamespace
from typing import Any, Awaitable, Callable, Dict, Generator, Iterable, Optional, Tuple, Type, Union, cast

from pjrpc import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, Request, Response, common
from pjrpc.client import retry
from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions, generators, v20
from pjrpc.common.typedefs import JsonRpcRequestId, MethodType

from .tracer import Tracer

logger = logging.getLogger(__package__)


class BaseBatch(abc.ABC):
    """
    Base batch wrapper. Implements some methods to wrap multiple JSON-RPC requests into a single batch request.

    :param client: JSON-RPC client instance
    """

    class BaseProxy(abc.ABC):
        """
        Proxy object. Provides syntactic sugar to make method calls using dot notation.

        :param batch: batch wrapper
        """

        def __init__(self, batch: 'BaseBatch'):
            self._batch = batch

        def __getattr__(self, attr: str) -> MethodType:
            def wrapped(*args: Any, **kwargs: Any) -> 'BaseBatch.BaseProxy':
                self._batch.add(attr, *args, **kwargs)
                return self

            return wrapped

        @abc.abstractmethod
        def __call__(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Union[Awaitable[Any], Any]:
            """
            Makes an RPC call.

            :param _trace_ctx: tracers request context
            """

        @abc.abstractmethod
        def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Union[Awaitable[Any], Any]:
            """
            Makes an RPC call.

            :param _trace_ctx: tracers request context
            """

    @property
    @abc.abstractmethod
    def proxy(self) -> 'BaseProxy':
        """
        Batch request proxy object.
        """

    def __init__(self, client: 'BaseAbstractClient'):
        self._client = client
        self._id_gen = client.id_gen_impl()
        self._requests = client.batch_request_class()

    def __getitem__(self, requests: Iterable[Tuple[Any]]) -> Union[Awaitable[Any], Any]:
        """
        Adds requests to the batch and makes a request.

        :param requests: requests to be added to the batch
        :returns: request results as a tuple
        """

        self._requests.extend([
            self._client.request_class(
                method=method,
                params=params,
                id=next(self._id_gen),
            ) for method, *params in requests  # type: ignore[var-annotated]
        ])
        return self.call()

    def __call__(self, method: str, *args: Any, **kwargs: Any) -> 'BaseBatch':
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        return self.add(method, *args, **kwargs)

    @abc.abstractmethod
    def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Union[Awaitable[Any], Any]:
        """
        Makes a JSON-RPC request.

        :param _trace_ctx: tracers request context
        :returns: request results as a tuple
        """

    @abc.abstractmethod
    def send(
        self, request: BatchRequest, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Union[Awaitable[Optional[BatchResponse]], Optional[BatchResponse]]:
        """
        Sends a JSON-RPC batch request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :returns: response instance
        """

    def add(self, method: str, *args: Any, **kwargs: Any) -> 'BaseBatch':
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(self._client.request_class(method, args or kwargs, id=next(self._id_gen)))
        return self

    def notify(self, method: str, *args: Any, **kwargs: Any) -> 'BaseBatch':
        """
        Adds a notification request to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(self._client.request_class(method, args or kwargs))
        return self

    def _relate(self, batch_request: BatchRequest, batch_response: BatchResponse) -> None:
        """
        Sets requests `related` field. if `strict` flag is ``True``
        checks that all requests have theirs corresponding responses

        :param batch_request: batch request
        :param batch_response: batch response
        """

        if batch_response.is_success:
            response_map = {response.id: response for response in batch_response if response.id is not None}

            for request in batch_request:
                if request.id is not None:
                    response = response_map.pop(request.id, None)
                    if response is None and self._client.strict:
                        raise exceptions.IdentityError(f"response '{request.id}' not found")
                    elif response is not None:
                        response.related = request

            if response_map and self._client.strict:
                raise exceptions.IdentityError(f"unexpected response found: {response_map.keys()}")


class Batch(BaseBatch):
    """
    Batch wrapper. Implements some methods to wrap multiple JSON-RPC requests into a single batch request.

    :param client: JSON-RPC client instance
    """

    class Proxy(BaseBatch.BaseProxy):

        def __init__(self, batch: 'Batch'):
            super().__init__(batch)

        def __call__(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            return self.call(_trace_ctx)

        def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            return self._batch.call(_trace_ctx)

    @property
    def proxy(self) -> 'Proxy':
        return Batch.Proxy(self)

    def __init__(self, client: 'AbstractClient'):
        super().__init__(client)

    def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Optional[Any]:
        response = self.send(self._requests)

        return response.result if response is not None else None

    def send(
        self, request: BatchRequest, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Optional[BatchResponse]:
        return cast(
            Optional[BatchResponse], self._client._send(
                request,
                response_class=self._client.batch_response_class,
                validator=self._relate,
                _trace_ctx=_trace_ctx,
                **kwargs,
            ),
        )


class AsyncBatch(BaseBatch):
    """
    Asynchronous batch wrapper. Used to make asynchronous JSON-RPC batch requests.
    """

    class Proxy(BaseBatch.BaseProxy):

        def __init__(self, batch: 'AsyncBatch'):
            super().__init__(batch)

        async def __call__(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            return await self.call(_trace_ctx)

        async def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            return await self._batch.call(_trace_ctx)

    @property
    def proxy(self) -> 'Proxy':
        return AsyncBatch.Proxy(self)

    def __init__(self, client: 'AbstractAsyncClient'):
        super().__init__(client)

    async def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Optional[Any]:
        response = await self.send(self._requests, _trace_ctx=_trace_ctx)

        return response.result if response is not None else None

    async def send(
        self, request: BatchRequest, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Optional[BatchResponse]:
        return await cast(
            Awaitable[Optional[BatchResponse]], self._client._send(
                request,
                response_class=self._client.batch_response_class,
                validator=self._relate,
                _trace_ctx=_trace_ctx,
                **kwargs,
            ),
        )


class BaseAbstractClient(abc.ABC):
    """
    Base abstract JSON-RPC client.

    :param request_class: request class
    :param response_class: response class
    :param batch_request_class: batch request class
    :param batch_response_class: batch response class
    :param error_cls: JSON-RPC error base class
    :param id_gen_impl: identifier generator
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    :param strict: if ``True`` checks that a request and a response identifiers match
    :param request_args: backend request argument
    :param tracers: request tracers list
    :param retry_strategy: request retry strategy
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param client: JSON-RPC client instance
        """

        def __init__(self, client: 'BaseAbstractClient'):
            self._client = client

        def __getattr__(self, attr: str) -> Callable[..., Any]:
            return ft.partial(self._client.call, attr)

    def __init__(
        self,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request_class: Type[BatchRequest] = v20.BatchRequest,
        batch_response_class: Type[BatchResponse] = v20.BatchResponse,
        error_cls: Type[exceptions.JsonRpcError] = exceptions.JsonRpcError,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestId, None, None]] = generators.sequential,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: Type[common.JSONEncoder] = common.JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        strict: bool = True,
        request_args: Optional[Dict[str, Any]] = None,
        tracers: Iterable[Tracer] = (),
        retry_strategy: Optional[retry.RetryStrategy] = None,
    ):
        self.request_class = request_class
        self.response_class = response_class
        self.batch_request_class = batch_request_class
        self.batch_response_class = batch_response_class
        self.error_cls = error_cls
        self.json_loader = json_loader
        self.json_dumper = json_dumper
        self.json_encoder = json_encoder
        self.json_decoder = json_decoder
        self.id_gen_impl = id_gen_impl
        self.strict = strict
        self._request_args = request_args or {}
        self._tracers = tracers
        self._retry_strategy = retry_strategy

    def __call__(
        self,
        method: str,
        *args: Any,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Union[Awaitable[Any], Any]:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        return self.call(method, *args, **kwargs)

    @property
    def proxy(self) -> 'Proxy':
        """
        Clint proxy object.
        """

        return BaseAbstractClient.Proxy(self)

    @abc.abstractmethod
    def call(self, method: str, *args: Any, **kwargs: Any) -> Union[Awaitable[Any], Any]:
        pass

    @abc.abstractmethod
    def _send(
        self,
        request: AbstractRequest,
        response_class: Type[AbstractResponse],
        validator: Callable[..., None],
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Union[Awaitable[Optional[AbstractResponse]], Optional[AbstractResponse]]:
        pass

    def _relate(self, request: Request, response: Response) -> None:
        """
        Checks the the request and the response identifiers match.

        :param request: request
        :param response: response
        """

        if self.strict and response.id is not None and response.id != request.id:
            raise exceptions.IdentityError(
                f"response id doesn't match the request one: expected {request.id}, got {response.id}",
            )

        response.related = request


class AbstractClient(BaseAbstractClient):
    """
    Abstract synchronous JSON-RPC client.
    """

    @property
    def batch(self) -> Batch:
        """
        Client batch wrapper.
        """

        return Batch(self)

    @abc.abstractmethod
    def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :returns: response text representation
        """

    def notify(
        self,
        method: str,
        *args: Any,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[Response]:
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=None,
            method=method,
            params=args or kwargs,
        )
        return self.send(request, _trace_ctx=_trace_ctx)

    def call(
        self, method: str, *args: Any, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Any:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = self.send(request, _trace_ctx=_trace_ctx)

        assert response is not None, "response is not set"
        return response.result

    def send(
        self,
        request: Request,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        _retry_strategy: MaybeSet[retry.RetryStrategy] = UNSET,
        **kwargs: Any,
    ) -> Optional[Response]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :param _retry_strategy: request retry strategy
        :returns: response instance
        """

        return cast(
            Response, self._send(
                request,
                response_class=self.response_class,
                validator=self._relate,
                _trace_ctx=_trace_ctx,
                _retry_strategy=_retry_strategy,
                **kwargs,
            ),
        )

    def traced(method: Callable[..., Any]) -> Callable[..., Any]:
        @ft.wraps(method)
        def wrapper(
            self: 'AbstractClient',
            request: AbstractRequest,
            _trace_ctx: SimpleNamespace,
            **kwargs: Any,
        ) -> Optional[AbstractResponse]:
            """
            Adds tracing logic to the method.
            """

            for tracer in self._tracers:
                tracer.on_request_begin(_trace_ctx, request)

            try:
                response = method(self, request, _trace_ctx=_trace_ctx, **kwargs)
            except BaseException as e:
                for tracer in self._tracers:
                    tracer.on_error(_trace_ctx, request, e)
                raise

            for tracer in self._tracers:
                tracer.on_request_end(_trace_ctx, request, response)

            return response

        return wrapper

    def retried(method: Callable[..., Any]) -> Callable[..., Any]:
        @ft.wraps(method)
        def wrapper(
            self: 'AbstractClient',
            request: AbstractRequest,
            _retry_strategy: MaybeSet[retry.RetryStrategy] = UNSET,
            **kwargs: Any,
        ) -> Optional[AbstractResponse]:
            """
            Adds retrying logic to the method.
            """

            retry_strategy = self._retry_strategy if isinstance(_retry_strategy, UnsetType) else _retry_strategy
            if retry_strategy:
                wrapped_method = retry.retry(method, retry_strategy)
            else:
                wrapped_method = method

            response = wrapped_method(self, request, **kwargs)

            return response

        return wrapper

    @retried
    @traced
    def _send(
        self,
        request: AbstractRequest,
        response_class: Type[AbstractResponse],
        validator: Callable[..., None],
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[AbstractResponse]:
        kwargs = {**self._request_args, **kwargs}
        request_text = self.json_dumper(request, cls=self.json_encoder)

        response_text = self._request(request_text, request.is_notification, **kwargs)
        if not request.is_notification:
            response = response_class.from_json(
                self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls,
            )
            validator(request, response)

        else:
            if self.strict and response_text:
                raise exceptions.BaseError("unexpected response")
            response = None

        return response


class AbstractAsyncClient(BaseAbstractClient):
    """
    Abstract asynchronous JSON-RPC client.
    """

    @property
    def batch(self) -> AsyncBatch:
        """
        Client batch wrapper.
        """

        return AsyncBatch(self)

    @abc.abstractmethod
    async def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :returns: response text representation
        """

    async def send(
        self,
        request: Request,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        _retry_strategy: MaybeSet[retry.RetryStrategy] = UNSET,
        **kwargs: Any,
    ) -> Optional[Response]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :param _retry_strategy: request retry strategy
        :returns: response instance
        """

        return cast(
            Response, await self._send(
                request,
                _trace_ctx=_trace_ctx,
                _retry_strategy=_retry_strategy,
                response_class=self.response_class,
                validator=self._relate,
                **kwargs,
            ),
        )

    def traced(method: Callable[..., Any]) -> Callable[..., Any]:
        @ft.wraps(method)
        async def wrapper(
            self: 'AbstractAsyncClient',
            request: Request,
            _trace_ctx: SimpleNamespace = SimpleNamespace(),
            **kwargs: Any,
        ) -> Response:
            """
            Adds tracing logic to the method.
            """

            for tracer in self._tracers:
                tracer.on_request_begin(_trace_ctx, request)

            try:
                response = await method(self, request, _trace_ctx=_trace_ctx, **kwargs)
            except BaseException as e:
                for tracer in self._tracers:
                    tracer.on_error(_trace_ctx, request, e)
                raise

            for tracer in self._tracers:
                tracer.on_request_end(_trace_ctx, request, response)

            return response

        return wrapper

    def retried(method: Callable[..., Awaitable[Any]]) -> Callable[..., Any]:
        @ft.wraps(method)
        async def wrapper(
            self: 'AbstractClient',
            request: AbstractRequest,
            _retry_strategy: MaybeSet[retry.RetryStrategy] = UNSET,
            **kwargs: Any,
        ) -> Optional[AbstractResponse]:
            """
            Adds retrying logic to the method.
            """

            retry_strategy = self._retry_strategy if isinstance(_retry_strategy, UnsetType) else _retry_strategy
            if retry_strategy:
                wrapped_method = retry.retry_async(method, retry_strategy)
            else:
                wrapped_method = method

            response = await wrapped_method(self, request, **kwargs)

            return response

        return wrapper

    @retried
    @traced
    async def _send(
        self,
        request: AbstractRequest,
        response_class: Type[AbstractResponse],
        validator: Callable[..., None],
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[AbstractResponse]:
        kwargs = {**self._request_args, **kwargs}
        request_text = self.json_dumper(request, cls=self.json_encoder)

        response_text = await self._request(request_text, request.is_notification, **kwargs)
        if not request.is_notification:
            response = response_class.from_json(
                self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls,
            )
            validator(request, response)

        else:
            if self.strict and response_text:
                raise exceptions.BaseError("unexpected response")
            response = None

        return response

    async def notify(
        self,
        method: str,
        *args: Any,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[Response]:
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=None,
            method=method,
            params=args or kwargs,
        )
        return await self.send(request, _trace_ctx=_trace_ctx)

    async def call(
        self, method: str, *args: Any, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Any:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = await self.send(request, _trace_ctx=_trace_ctx)

        assert response is not None, "response is not set"
        return response.result



"""
Builtin request id generators. Implements several identifier types and generation strategies.
"""

import itertools as it
import random as _random
import string
import uuid as _uuid
from typing import Generator


def sequential(start: int = 1, step: int = 1) -> Generator[int, None, None]:
    """
    Sequential id generator. Returns consecutive values starting from `start` with step `step`.

    :param start: starting number
    :param step: step
    """

    yield from it.count(start, step)


def randint(a: int, b: int) -> Generator[int, None, None]:
    """
    Random integer id generator. Returns random integers between `a` and `b`.

    :param a: from
    :param b: to
    """

    while True:
        yield _random.randint(a, b)


def random(length: int = 8, chars: str = string.digits + string.ascii_lowercase) -> Generator[str, None, None]:
    """
    Random string id generator. Returns random strings of length `length` using alphabet `chars`.

    :param length: string length
    :param chars: string characters
    """

    while True:
        yield ''.join((_random.choice(chars) for _ in range(length)))


def uuid() -> Generator[_uuid.UUID, None, None]:
    """
    UUID id generator. Returns random UUIDs.
    """

    while True:
        yield _uuid.uuid4()



"""
Definition of package exceptions and JSON-RPC protocol errors.
"""

import typing
from typing import Any, Dict, Optional, Tuple, Type

from pjrpc.common.typedefs import Json

from .common import UNSET, MaybeSet


class BaseError(Exception):
    """
    Base package error. All package errors are inherited from it.
    """


class IdentityError(BaseError):
    """
    Raised when a batch requests/responses identifiers are not unique or missing.
    """


class DeserializationError(BaseError, ValueError):
    """
    Request/response deserializatoin error.
    Raised when request/response json has incorrect format.
    """


class JsonRpcErrorMeta(type):
    """
    :py:class:`pjrpc.common.exceptions.JsonRpcError` metaclass.
    Builds a mapping from an error code number to an error class
    inherited from a :py:class:`pjrpc.common.exceptions.JsonRpcError`.
    """

    __errors_mapping__: Dict[int, Type['JsonRpcError']] = {}

    def __new__(mcs, name: str, bases: Tuple[type, ...], dct: Dict[str, Any]) -> Type['JsonRpcError']:
        cls: Type['JsonRpcError'] = typing.cast(Type['JsonRpcError'], super().__new__(mcs, name, bases, dct))
        if hasattr(cls, 'code') and cls.code is not None:
            mcs.__errors_mapping__[cls.code] = cls

        return cls


class JsonRpcError(BaseError, metaclass=JsonRpcErrorMeta):
    """
    `JSON-RPC <https://www.jsonrpc.org>`_ protocol error.
    For more information see `Error object <https://www.jsonrpc.org/specification#error_object>`_.
    All JSON-RPC protocol errors are inherited from it.

    :param code: number that indicates the error type
    :param message: short description of the error
    :param data: value that contains additional information about the error. May be omitted.
    """

    # a number that indicates the error type that occurred
    code: int = None  # type: ignore[assignment]

    # a string providing a short description of the error.
    # the message SHOULD be limited to a concise single sentence.
    message: str = None  # type: ignore[assignment]

    @classmethod
    def from_json(cls, json_data: 'Json') -> 'JsonRpcError':
        """
        Deserializes an error from json data. If data format is not correct :py:class:`ValueError` is raised.

        :param json_data: json data the error to be deserialized from

        :returns: deserialized error
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(json_data, dict):
                raise DeserializationError("data must be of type dict")

            code = json_data['code']
            if not isinstance(code, int):
                raise DeserializationError("field 'code' must be of type integer")

            message = json_data['message']
            if not isinstance(message, str):
                raise DeserializationError("field 'message' must be of type string")

            error_class = cls.get_error_cls(code, cls)

            return error_class(code, message, json_data.get('data', UNSET))
        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    @classmethod
    def get_error_cls(cls, code: int, default: Type['JsonRpcError']) -> Type['JsonRpcError']:
        return type(cls).__errors_mapping__.get(code, default)

    def __init__(self, code: Optional[int] = None, message: Optional[str] = None, data: MaybeSet[Any] = UNSET):
        assert code or self.code, "code is not provided"
        assert message or self.message, "message is not provided"

        self.code = code or self.code
        self.message = message or self.message
        self.data = data

        super().__init__(code, message)

    def __str__(self) -> str:
        return "({code}) {message}".format(code=self.code, message=self.message)

    def __repr__(self) -> str:
        return "{class_name}(code={code}, message={message}, data={data})".format(
            class_name=self.__class__.__name__, code=repr(self.code), message=repr(self.message), data=repr(self.data),
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, JsonRpcError):
            return NotImplemented

        return (self.code, self.message, self.data) == (other.code, other.message, other.data)

    def to_json(self) -> 'Json':
        """
        Serializes the error to a dict.

        :returns: serialized error
        """

        json: Dict[str, Any] = {
            'code': self.code,
            'message': self.message,
        }
        if self.data is not UNSET:
            json.update(data=self.data)

        return json


class ClientError(JsonRpcError):
    """
    Raised when a client sent an incorrect request.
    """


class ParseError(ClientError):
    """
    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.
    """

    code = -32700
    message = 'Parse error'


class InvalidRequestError(ClientError):
    """
    The JSON sent is not a valid request object.
    """

    code = -32600
    message = 'Invalid Request'


class MethodNotFoundError(ClientError):
    """
    The method does not exist / is not available.
    """

    code = -32601
    message = 'Method not found'


class InvalidParamsError(ClientError):
    """
    Invalid method parameter(s).
    """

    code = -32602
    message = 'Invalid params'


class InternalError(JsonRpcError):
    """
    Internal JSON-RPC error.
    """

    code = -32603
    message = 'Internal error'


class ServerError(JsonRpcError):
    """
    Reserved for implementation-defined server-errors.
    Codes from -32000 to -32099.
    """

    code = -32000
    message = 'Server error'



import asyncio
import dataclasses as dc
import itertools as it
import logging
import time
from typing import Any, Awaitable, Callable, Generator, Iterator, Optional, Set, Type

from pjrpc.common import AbstractResponse

logger = logging.getLogger(__package__)

Jitter = Callable[[], float]


@dc.dataclass(frozen=True)
class Backoff:
    """
    JSON-RPC request retry strategy.

    :param attempts: retries number
    :param jitter: retry delay jitter generator
    """

    attempts: int
    jitter: Jitter = lambda: 0.0

    def __call__(self) -> Iterator[float]:
        """
        Returns delay iterator.
        """

        raise NotImplementedError


@dc.dataclass(frozen=True)
class PeriodicBackoff(Backoff):
    """
    Periodic request retry strategy.

    :param interval: retry delay
    """

    interval: float = 1.0

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            for _ in range(self.attempts):
                yield self.interval + self.jitter()

        return gen()


@dc.dataclass(frozen=True)
class ExponentialBackoff(Backoff):
    """
    Exponential request retry strategy.

    :param base: exponentially growing delay base
    :param factor: exponentially growing delay factor (multiplier)
    :param max_value: delay max value
    """

    base: float = 1.0
    factor: float = 2.0
    max_value: Optional[float] = None

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            for n, base in enumerate(it.repeat(self.base, self.attempts)):
                value = base * (self.factor ** n) + self.jitter()
                yield min(self.max_value, value) if self.max_value is not None else value

        return gen()


@dc.dataclass(frozen=True)
class FibonacciBackoff(Backoff):
    """
    Fibonacci request retry strategy.

    :param multiplier: fibonacci interval sequence multiplier
    :param max_value: delay max value
    """

    multiplier: float = 1.0
    max_value: float = 1.0

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            prev, cur = 1, 1

            for _ in range(self.attempts):
                value = cur * self.multiplier + self.jitter()
                yield min(self.max_value, value) if self.max_value is not None else value

                tmp = cur
                cur = prev + cur
                prev = tmp

        return gen()


@dc.dataclass(frozen=True)
class RetryStrategy:
    """
    JSON-RPC request retry strategy.

    :param backoff: backoff delay generator
    :param codes: JSON-RPC response codes receiving which the request will be retried
    :param exceptions: exceptions catching which the request will be retried
    """

    backoff: Backoff
    codes: Optional[Set[int]] = None
    exceptions: Optional[Set[Type[Exception]]] = None


def retry(
    func: Callable[..., AbstractResponse],
    retry_strategy: RetryStrategy,
) -> Callable[..., AbstractResponse]:
    """
    Synchronous function retry decorator.

    :param func: function to be retried
    :param retry_strategy: retry strategy to be applied
    :return: decorated function
    """

    def wrapped(*args: Any, **kwargs: Any) -> AbstractResponse:
        delays = retry_strategy.backoff()

        for attempt in it.count(start=1):
            try:
                response = func(*args, **kwargs)
                if response.is_error and retry_strategy.codes and response.get_error().code in retry_strategy.codes:
                    delay = next(delays, None)
                    if delay is not None:
                        logger.debug("retrying request: attempt=%d, code=%s", attempt, response.error)
                        time.sleep(delay)
                        continue

                return response

            except tuple(retry_strategy.exceptions or {}) as e:
                delay = next(delays, None)
                if delay is not None:
                    logger.debug("retrying request: attempt=%d, exception=%r", attempt, e)
                    time.sleep(delay)
                else:
                    raise e
        else:
            raise AssertionError("unreachable")

    return wrapped


def retry_async(
    func: Callable[..., Awaitable[AbstractResponse]],
    retry_strategy: RetryStrategy,
) -> Callable[..., Awaitable[AbstractResponse]]:
    """
    Asynchronous function retry decorator.

    :param func: function to be retried
    :param retry_strategy: retry strategy to be applied
    :return: decorated function
    """

    async def wrapped(*args: Any, **kwargs: Any) -> AbstractResponse:
        delays = retry_strategy.backoff()

        for attempt in it.count(start=1):
            try:
                response = await func(*args, **kwargs)
                if response.is_error and retry_strategy.codes and response.get_error().code in retry_strategy.codes:
                    delay = next(delays, None)
                    if delay is not None:
                        logger.debug("retrying request: attempt=%d, code=%s", attempt, response.error)
                        await asyncio.sleep(delay)
                        continue

                return response

            except tuple(retry_strategy.exceptions or {}) as e:
                delay = next(delays, None)
                if delay is not None:
                    logger.debug("retrying request: attempt=%d, exception=%r", attempt, e)
                    await asyncio.sleep(delay)
                else:
                    raise e
        else:
            raise AssertionError("unreachable")

    return wrapped



import json
import sys
from typing import Any, Dict, TypeVar, Union

if sys.version_info >= (3, 8):
    from typing import Literal
    FalseType = Literal[False]
else:
    FalseType = bool

import pjrpc
from pjrpc.common.typedefs import Json  # noqa: for back compatibility


class UnsetType:
    """
    `Sentinel <https://en.wikipedia.org/wiki/Sentinel_value>`_ object.
    Used to distinct unset (missing) values from ``None`` ones.
    """

    def __bool__(self) -> FalseType:
        return False

    def __repr__(self) -> str:
        return "UNSET"

    def __str__(self) -> str:
        return repr(self)

    def __copy__(self) -> 'UnsetType':
        return self

    def __deepcopy__(self, memo: Dict[str, Any]) -> 'UnsetType':
        return self


UNSET: UnsetType = UnsetType()

MaybeSetType = TypeVar('MaybeSetType')
MaybeSet = Union[UnsetType, MaybeSetType]


class JSONEncoder(json.JSONEncoder):
    """
    Library default JSON encoder. Encodes request, response and error objects to be json serializable.
    All custom encoders should be inherited from it.
    """

    def default(self, o: Any) -> Any:
        if isinstance(
            o, (
                pjrpc.Response, pjrpc.Request,
                pjrpc.BatchResponse, pjrpc.BatchRequest,
                pjrpc.exceptions.JsonRpcError,
            ),
        ):
            return o.to_json()

        return super().default(o)



import logging
from typing import Any, Dict, Optional

import aio_pika
from yarl import URL

import pjrpc.server

logger = logging.getLogger(__package__)


class Executor:
    """
    `aio_pika <https://aio-pika.readthedocs.io/en/latest/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param queue_name: requests queue name
    :param prefetch_count: worker prefetch count
    :param kwargs: dispatcher additional arguments
    """

    def __init__(self, broker_url: URL, queue_name: str, prefetch_count: int = 0, **kwargs: Any):
        self._broker_url = broker_url
        self._queue_name = queue_name
        self._prefetch_count = prefetch_count

        self._connection = aio_pika.connection.Connection(broker_url)
        self._channel: Optional[aio_pika.abc.AbstractChannel] = None

        self._queue: Optional[aio_pika.abc.AbstractQueue] = None
        self._consumer_tag: Optional[str] = None

        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)

    @property
    def dispatcher(self) -> pjrpc.server.AsyncDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def start(self, queue_args: Optional[Dict[str, Any]] = None) -> None:
        """
        Starts executor.

        :param queue_args: queue arguments
        """

        await self._connection.connect()
        self._channel = channel = await self._connection.channel()

        self._queue = queue = await channel.declare_queue(self._queue_name, **(queue_args or {}))
        await channel.set_qos(prefetch_count=self._prefetch_count)
        self._consumer_tag = await queue.consume(self._rpc_handle)

    async def shutdown(self) -> None:
        """
        Stops executor.
        """

        if self._consumer_tag and self._queue:
            await self._queue.cancel(self._consumer_tag)
        if self._channel:
            await self._channel.close()

        await self._connection.close()

    async def _rpc_handle(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        """
        Handles JSON-RPC request.

        :param message: incoming message
        """

        try:
            reply_to = message.reply_to
            response_text = await self._dispatcher.dispatch(message.body.decode(), context=message)

            if response_text is not None:
                if reply_to is None:
                    logger.warning("property 'reply_to' is missing")
                else:
                    async with self._connection.channel() as channel:
                        await channel.default_exchange.publish(
                            aio_pika.Message(
                                body=response_text.encode(),
                                reply_to=reply_to,
                                correlation_id=message.correlation_id,
                                content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                            ),
                            routing_key=reply_to,
                        )

            await message.ack()

        except Exception as e:
            logger.exception("jsonrpc request handling error: %s", e)



import logging
from types import SimpleNamespace
from typing import Optional

from pjrpc import AbstractRequest, AbstractResponse

client_logger = logging.getLogger(__package__)


class Tracer:
    """
    JSON-RPC client tracer.
    """

    def on_request_begin(self, trace_context: SimpleNamespace, request: AbstractRequest) -> None:
        """
        Handler called before JSON-RPC request begins.

        :param trace_context: request trace context
        :param request: JSON-RPC request
        """

    def on_request_end(
        self, trace_context: SimpleNamespace, request: AbstractRequest, response: Optional[AbstractResponse],
    ) -> None:
        """
        Handler called after JSON-RPC request ends.

        :param trace_context: request trace context
        :param request: JSON-RPC request
        :param response: JSON-RPC response
        """

    def on_error(
        self, trace_context: SimpleNamespace, request: AbstractRequest, error: BaseException,
    ) -> None:
        """
        Handler called when JSON-RPC request failed.

        :param trace_context: request trace context
        :param request: JSON-RPC request
        :param error: raised exception
        """


class LoggingTracer(Tracer):
    """
    JSON-RPC client logging tracer.
    """

    def __init__(self, logger: logging.Logger = client_logger, level: int = logging.DEBUG):
        self._logger = logger
        self._level = level

    def on_request_begin(self, trace_context: SimpleNamespace, request: AbstractRequest) -> None:
        self._logger.log(self._level, "sending request: %r", request)

    def on_request_end(
        self, trace_context: SimpleNamespace, request: AbstractRequest, response: Optional[AbstractResponse],
    ) -> None:
        self._logger.log(self._level, "received response: %r", response)

    def on_error(
        self, trace_context: SimpleNamespace, request: AbstractRequest, error: BaseException,
    ) -> None:
        self._logger.log(self._level, "request '%s' sending error: %r", request, error)



"""
Flask JSON-RPC extension.
"""

import functools as ft
import json
from typing import Any, Dict, Optional, Union

import flask
from flask import current_app
from werkzeug import exceptions

import pjrpc.server
from pjrpc.server import specs, utils


class JsonRPC:
    """
    `Flask <https://flask.palletsprojects.com/en/1.1.x/>`_ framework JSON-RPC extension class.

    :param path: JSON-RPC handler base path
    :param spec: JSON-RPC specification
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
    """

    def __init__(self, path: str, spec: Optional[specs.Specification] = None, **kwargs: Any):
        self._path = path.rstrip('/')
        self._spec = spec

        kwargs.setdefault('json_loader', flask.json.loads)
        kwargs.setdefault('json_dumper', flask.json.dumps)

        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)
        self._endpoints: Dict[str, pjrpc.server.Dispatcher] = {'': self._dispatcher}
        self._blueprints: Dict[str, flask.Blueprint] = {}

    @property
    def dispatcher(self) -> pjrpc.server.Dispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def endpoints(self) -> Dict[str, pjrpc.server.Dispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_endpoint(
        self,
        prefix: str,
        blueprint: Optional[flask.Blueprint] = None,
        **kwargs: Any,
    ) -> pjrpc.server.Dispatcher:
        """
        Adds additional endpoint.

        :param prefix: endpoint prefix
        :param blueprint: flask blueprint the endpoint will be served on
        :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
        :return: dispatcher
        """

        prefix = prefix.rstrip('/')
        dispatcher = pjrpc.server.Dispatcher(**kwargs)

        self._endpoints[prefix] = dispatcher
        if blueprint is not None:
            self._blueprints[prefix] = blueprint

        return dispatcher

    def init_app(self, app: Union[flask.Flask, flask.Blueprint]) -> None:
        """
        Initializes flask application with JSON-RPC extension.

        :param app: flask application instance
        """

        for prefix, dispatcher in self._endpoints.items():
            path = utils.join_path(self._path, prefix)
            blueprint = self._blueprints.get(prefix)

            (blueprint or app).add_url_rule(
                path,
                methods=['POST'],
                view_func=ft.partial(self._rpc_handle, dispatcher=dispatcher),
                endpoint=path.replace('/', '_'),
            )
            if blueprint:
                app.register_blueprint(blueprint)

        if self._spec:
            app.add_url_rule(
                utils.join_path(self._path, self._spec.path),
                methods=['GET'],
                view_func=self._generate_spec,
            )

            if self._spec.ui and self._spec.ui_path:
                path = utils.join_path(self._path, self._spec.ui_path)
                app.add_url_rule(f'{path}/', methods=['GET'], view_func=self._ui_index_page)
                app.add_url_rule(f'{path}/index.html', methods=['GET'], view_func=self._ui_index_page)
                app.add_url_rule(f'{path}/<path:filename>', methods=['GET'], view_func=self._ui_static)

    def _generate_spec(self) -> flask.Response:
        assert self._spec is not None, "spec is not set"

        endpoint_path = utils.remove_suffix(flask.request.path, suffix=self._spec.path)
        methods = {path: dispatcher.registry.values() for path, dispatcher in self._endpoints.items()}
        schema = self._spec.schema(path=endpoint_path, methods_map=methods)

        return current_app.response_class(
            json.dumps(schema, cls=specs.JSONEncoder),
            mimetype=pjrpc.common.DEFAULT_CONTENT_TYPE,
        )

    def _ui_index_page(self) -> flask.Response:
        assert self._spec is not None and self._spec.ui is not None, "spec is not set"

        app_path = flask.request.path.rsplit(self._spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, self._spec.path)

        return current_app.response_class(
            response=self._spec.ui.get_index_page(spec_url=spec_full_path),
            content_type='text/html',
        )

    def _ui_static(self, filename: str) -> flask.Response:
        assert self._spec is not None and self._spec.ui is not None, "spec is not set"

        return flask.send_from_directory(self._spec.ui.get_static_folder(), filename)

    def _rpc_handle(self, dispatcher: pjrpc.server.Dispatcher) -> flask.Response:
        """
        Handles JSON-RPC request.

        :returns: flask response
        """

        if not flask.request.is_json:
            raise exceptions.UnsupportedMediaType()

        try:
            flask.request.encoding_errors = 'strict'
            request_text = flask.request.get_data(as_text=True)
        except UnicodeDecodeError as e:
            raise exceptions.BadRequest() from e

        response_text = dispatcher.dispatch(request_text)
        if response_text is None:
            return current_app.response_class()
        else:
            return current_app.response_class(response_text, mimetype=pjrpc.common.DEFAULT_CONTENT_TYPE)



"""
aiohttp JSON-RPC server integration.
"""

import functools as ft
import json
from typing import Any, Dict, Mapping, Optional, cast

from starlette import exceptions, routing
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.staticfiles import StaticFiles

import pjrpc
from pjrpc.common.typedefs import Func
from pjrpc.server import specs, utils


def async_partial(func: Func, *bound_args: Any, **bound_kwargs: Any) -> Func:
    @ft.wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        return await func(*bound_args, *args, **bound_kwargs, **kwargs)

    return cast(Func, wrapped)


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param app_args: arguments to be passed to :py:class:`aiohttp.web.Application`
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(
        self,
        path: str = '',
        spec: Optional[specs.Specification] = None,
        app: Optional[Starlette] = None,
        **kwargs: Any,
    ):
        self._path = path.rstrip('/')
        self._spec = spec
        self._app = app or Starlette()
        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)
        self._endpoints: Dict[str, pjrpc.server.AsyncDispatcher] = {'': self._dispatcher}

        self._app.add_route(self._path, async_partial(self._rpc_handle, dispatcher=self._dispatcher), methods=['POST'])

        if self._spec:
            self._app.add_route(utils.join_path(self._path, self._spec.path), self._generate_spec, methods=['GET'])

            if self._spec.ui and self._spec.ui_path:
                ui_path = utils.join_path(self._path, self._spec.ui_path)

                self._app.add_route(utils.join_path(ui_path, '/'), self._ui_index_page)
                self._app.add_route(utils.join_path(ui_path, 'index.html'), self._ui_index_page)
                self._app.routes.append(
                    routing.Mount(ui_path, app=StaticFiles(directory=str(self._spec.ui.get_static_folder()))),
                )

    @property
    def app(self) -> Starlette:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> pjrpc.server.AsyncDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def endpoints(self) -> Mapping[str, pjrpc.server.AsyncDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_endpoint(self, prefix: str, **kwargs: Any) -> pjrpc.server.AsyncDispatcher:
        """
        Adds additional endpoint.

        :param prefix: endpoint prefix
        :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
        :return: dispatcher
        """

        dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)
        self._endpoints[prefix] = dispatcher

        self._app.add_route(
            utils.join_path(self._path, prefix),
            async_partial(self._rpc_handle, dispatcher=self._dispatcher),
            methods=['POST'],
        )

        return dispatcher

    async def _generate_spec(self, request: Request) -> Response:
        assert self._spec is not None, "spec is not set"

        endpoint_path = utils.remove_suffix(request.url.path, suffix=self._spec.path)

        methods = {path: dispatcher.registry.values() for path, dispatcher in self._endpoints.items()}
        schema = self._spec.schema(path=endpoint_path, methods_map=methods)

        return Response(
            content=json.dumps(schema, cls=specs.JSONEncoder),
            media_type='application/json',
        )

    async def _ui_index_page(self, request: Request) -> Response:
        assert self._spec is not None and self._spec.ui is not None, "spec is not set"

        app_path = request.url.path.rsplit(self._spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, self._spec.path)

        return Response(
            content=self._spec.ui.get_index_page(spec_url=spec_full_path),
            media_type='text/html',
        )

    async def _rpc_handle(self, http_request: Request, dispatcher: pjrpc.server.AsyncDispatcher) -> Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.headers['Content-Type'] not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise exceptions.HTTPException(415)

        try:
            request_data = await http_request.body()
            request_text = request_data.decode()
        except UnicodeDecodeError as e:
            raise exceptions.HTTPException(400) from e

        response_text = await dispatcher.dispatch(request_text, context=http_request)
        if response_text is None:
            return Response()
        else:
            return Response(content=response_text, media_type=pjrpc.common.DEFAULT_CONTENT_TYPE)



from typing import Any, Callable, Dict, Iterable, TypeVar


def get_meta(instance: Any) -> Dict[str, Any]:
    """
    Returns object pjrpc metadata.
    """

    return getattr(instance, '__pjrpc_meta__', {})


def set_meta(instance: Any, **meta: Any) -> Dict[str, Any]:
    """
    Updates object pjrpc metadata.
    """

    if not hasattr(instance, '__pjrpc_meta__'):
        instance.__pjrpc_meta__ = {}

    instance.__pjrpc_meta__.update(meta)

    return instance.__pjrpc_meta__


def remove_prefix(s: str, prefix: str) -> str:
    """
    Removes a prefix from a string.

    :param s: string to be processed
    :param prefix: prefix to be removed
    :return: processed string
    """

    if s.startswith(prefix):
        return s[len(prefix):]
    else:
        return s


def remove_suffix(s: str, suffix: str) -> str:
    """
    Removes a suffix from a string.

    :param s: string to be processed
    :param suffix: suffix to be removed
    :return: processed string
    """

    if suffix and s.endswith(suffix):
        return s[0:-len(suffix)]
    else:
        return s


def join_path(path: str, *paths: str) -> str:
    result = path
    for path in paths:
        if path:
            result = f'{result.rstrip("/")}/{path.lstrip("/")}'

    return result


T = TypeVar('T')
K = TypeVar('K')


def unique(*iterables: Iterable[T], key: Callable[[T], K]) -> Iterable[T]:
    items_map: Dict[K, T] = {}
    for iterable in iterables:
        for item in iterable:
            items_map[key(item)] = item

    return items_map.values()



"""
aiohttp JSON-RPC server integration.
"""

import functools as ft
import json
from typing import Any, Dict, Mapping, Optional

import aiohttp.web
from aiohttp import web

import pjrpc
from pjrpc.server import specs, utils


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param app_args: arguments to be passed to :py:class:`aiohttp.web.Application`
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(
        self,
        path: str = '',
        spec: Optional[specs.Specification] = None,
        app: Optional[web.Application] = None,
        **kwargs: Any,
    ):
        self._path = path.rstrip('/')
        self._spec = spec
        self._app = app or web.Application()
        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)
        self._endpoints: Dict[str, pjrpc.server.AsyncDispatcher] = {'': self._dispatcher}

        self._app.router.add_post(self._path, ft.partial(self._rpc_handle, dispatcher=self._dispatcher))

        if self._spec:
            self._app.router.add_get(utils.join_path(self._path, self._spec.path), self._generate_spec)

            if self._spec.ui and self._spec.ui_path:
                ui_app = web.Application()
                ui_app.router.add_get('/', self._ui_index_page)
                ui_app.router.add_get('/index.html', self._ui_index_page)
                ui_app.router.add_static('/', self._spec.ui.get_static_folder())

                self._app.add_subapp(utils.join_path(self._path, self._spec.ui_path), ui_app)

    @property
    def app(self) -> web.Application:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> pjrpc.server.AsyncDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def endpoints(self) -> Mapping[str, pjrpc.server.AsyncDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_endpoint(
        self,
        prefix: str,
        subapp: Optional[aiohttp.web.Application] = None,
        **kwargs: Any,
    ) -> pjrpc.server.AsyncDispatcher:
        """
        Adds additional endpoint.

        :param prefix: endpoint prefix
        :param subapp: aiohttp subapp the endpoint will be served on
        :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
        :return: dispatcher
        """

        prefix = prefix.rstrip('/')
        dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)
        self._endpoints[prefix] = dispatcher

        if subapp:
            subapp.router.add_post('', ft.partial(self._rpc_handle, dispatcher=dispatcher))
            self._app.add_subapp(utils.join_path(self._path, prefix), subapp)
        else:
            self._app.router.add_post(
                utils.join_path(self._path, prefix),
                ft.partial(self._rpc_handle, dispatcher=dispatcher),
            )

        return dispatcher

    async def _generate_spec(self, request: web.Request) -> web.Response:
        assert self._spec is not None, "spec is not set"

        endpoint_path = utils.remove_suffix(request.path, suffix=self._spec.path)

        methods = {path: dispatcher.registry.values() for path, dispatcher in self._endpoints.items()}
        schema = self._spec.schema(path=endpoint_path, methods_map=methods)

        return web.json_response(text=json.dumps(schema, cls=specs.JSONEncoder))

    async def _ui_index_page(self, request: web.Request) -> web.Response:
        assert self._spec is not None and self._spec.ui is not None, "spec is not set"

        app_path = request.path.rsplit(self._spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, self._spec.path)

        return web.Response(
            text=self._spec.ui.get_index_page(spec_url=spec_full_path),
            content_type='text/html',
        )

    async def _rpc_handle(self, http_request: web.Request, dispatcher: pjrpc.server.AsyncDispatcher) -> web.Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response_text = await dispatcher.dispatch(request_text, context=http_request)
        if response_text is None:
            return web.Response()
        else:
            return web.json_response(text=response_text)



"""
JSON-RPC client.
"""

from .client import AbstractAsyncClient, AbstractClient, AsyncBatch, BaseAbstractClient, BaseBatch, Batch
from .tracer import LoggingTracer, Tracer

__all__ = [
    'AbstractClient',
    'AbstractAsyncClient',
    'AsyncBatch',
    'BaseBatch',
    'BaseAbstractClient',
    'Batch',
    'LoggingTracer',
    'Tracer',
]



from typing import Any, Optional

import requests

import pjrpc
from pjrpc.client import AbstractClient


class Client(AbstractClient):
    """
    `Requests <https://2.python-requests.org/>`_ library client backend.

    :param url: url to be used as JSON-RPC endpoint.
    :param session: custom session to be used instead of :py:class:`requests.Session`
    :param kwargs: parameters to be passed to :py:class:`pjrpc.client.AbstractClient`
    """

    def __init__(self, url: str, session: Optional[requests.Session] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._endpoint = url
        self._session = session or requests.Session()

    def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param data: request text
        :param is_notification: is the request a notification
        :returns: response text
        """

        kwargs = {
            'headers': {'Content-Type': pjrpc.common.DEFAULT_CONTENT_TYPE},
            **kwargs,
        }

        resp = self._session.post(self._endpoint, data=request_text, **kwargs)
        resp.raise_for_status()
        if is_notification:
            return None

        response_text = resp.text
        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in pjrpc.common.RESPONSE_CONTENT_TYPES:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    def close(self) -> None:
        """
        Closes the current http session.
        """

        self._session.close()

    def __enter__(self) -> 'Client':
        self._session.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        self._session.__exit__(*args)



import functools as ft
import inspect
from typing import Any, Dict, Iterable, Optional, Tuple

from pjrpc.common.typedefs import JsonRpcParams, MethodType
from pjrpc.server import utils


class ValidationError(Exception):
    """
    Method parameters validation error. Raised when parameters validation failed.
    """


class BaseValidator:
    """
    Base method parameters validator. Uses :py:func:`inspect.signature` for validation.
    """

    def validate(self, maybe_method: Optional[MethodType] = None, **kwargs: Any) -> MethodType:
        """
        Decorator marks a method the parameters of which to be validated when calling it using JSON-RPC protocol.

        :param maybe_method: method the parameters of which to be validated or ``None`` if called as @validate(...)
        :param kwargs: validator arguments
        """

        def decorator(method: MethodType) -> MethodType:
            utils.set_meta(method, validator=self, validator_args=kwargs)
            return method

        # maybe_method's type depends on the usage of the decorator.  It's a function
        # if it's used as `@validate` but ``None`` if used as `@validate()`.
        if maybe_method is None:
            return decorator
        else:
            return decorator(maybe_method)

    def validate_method(
        self, method: MethodType, params: Optional['JsonRpcParams'], exclude: Iterable[str] = (), **kwargs: Any,
    ) -> Dict[str, Any]:
        """
        Validates params against method signature.

        :param method: method to validate parameters against
        :param params: parameters to be validated
        :param exclude: parameter names to be excluded from validation
        :param kwargs: additional validator arguments

        :raises: :py:class:`pjrpc.server.validators.ValidationError`
        :returns: bound method parameters
        """

        signature = self.signature(method, tuple(exclude))
        return self.bind(signature, params).arguments

    def bind(self, signature: inspect.Signature, params: Optional['JsonRpcParams']) -> inspect.BoundArguments:
        """
        Binds parameters to method.
        :param signature: method to bind parameters to
        :param params: parameters to be bound

        :raises: ValidationError is parameters binding failed
        :returns: bound parameters
        """

        method_args = params if isinstance(params, (list, tuple)) else ()
        method_kwargs = params if isinstance(params, dict) else {}

        try:
            return signature.bind(*method_args, **method_kwargs)
        except TypeError as e:
            raise ValidationError(str(e)) from e

    @ft.lru_cache(None)
    def signature(self, method: MethodType, exclude: Tuple[str]) -> inspect.Signature:
        """
        Returns method signature.

        :param method: method to get signature of
        :param exclude: parameters to be excluded
        :returns: signature
        """

        signature = inspect.signature(method)
        parameters = signature.parameters.copy()
        for item in exclude:
            parameters.pop(item, None)

        return signature.replace(parameters=tuple(parameters.values()))



from typing import Dict, Iterable, List, Optional, Type

import docstring_parser

from pjrpc.common import UNSET, MaybeSet, exceptions
from pjrpc.common.typedefs import MethodType
from pjrpc.server.specs.extractors import BaseSchemaExtractor, Error, Example, JsonRpcError, Schema, Tag


class DocstringSchemaExtractor(BaseSchemaExtractor):
    """
    docstring method specification generator.
    """

    def extract_params_schema(self, method: MethodType, exclude: Iterable[str] = ()) -> Dict[str, Schema]:
        exclude = set(exclude)
        parameters_schema = {}

        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            for param in doc.params:
                if param.arg_name in exclude:
                    continue

                parameters_schema[param.arg_name] = Schema(
                    schema={'type': param.type_name},
                    required=not param.is_optional,
                    summary=param.description.split('.')[0] if param.description is not None else UNSET,
                    description=param.description if param.description is not None else UNSET,
                )

        return parameters_schema

    def extract_result_schema(self, method: MethodType) -> Schema:
        result_schema = Schema(schema={})

        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            if doc and doc.returns:
                result_schema = Schema(
                    schema={'type': doc.returns.type_name},
                    required=True,
                    summary=doc.returns.description.split('.')[0] if doc.returns.description is not None else UNSET,
                    description=doc.returns.description if doc.returns.description is not None else UNSET,
                )

        return result_schema

    def extract_errors_schema(
        self,
        method: MethodType,
        errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> MaybeSet[List[Error]]:
        errors_schema = []

        if method.__doc__:
            error_map = {
                error.__name__: error
                for error in exceptions.JsonRpcErrorMeta.__errors_mapping__.values()
            }

            doc = docstring_parser.parse(method.__doc__)
            for error in doc.raises:
                error_cls = error_map.get(error.type_name or '')
                if error_cls:
                    errors_schema.append(
                        Error(
                            code=error_cls.code,
                            message=error_cls.message,
                        ),
                    )

        return errors_schema or UNSET

    def extract_description(self, method: MethodType) -> MaybeSet[str]:
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            description = doc.long_description or UNSET
        else:
            description = UNSET

        return description

    def extract_summary(self, method: MethodType) -> MaybeSet[str]:
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            description = doc.short_description or UNSET
        else:
            description = UNSET

        return description

    def extract_tags(self, method: MethodType) -> MaybeSet[List[Tag]]:
        return UNSET

    def extract_examples(self, method: MethodType) -> MaybeSet[List[Example]]:
        return UNSET

    def extract_deprecation_status(self, method: MethodType) -> MaybeSet[bool]:
        is_deprecated: MaybeSet[bool]
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            is_deprecated = bool(doc.deprecation)
        else:
            is_deprecated = UNSET

        return is_deprecated



from __future__ import annotations

from typing import Any, Callable, Dict, List, Set, Tuple, TypeVar, Union

__all__ = [
    'JsonRpcParams',
    'JsonRpcRequestId',
    'Json',
    'MethodType',
    'Func',
]

JsonRpcParams = Union[List[Any], Tuple[Any, ...], Dict[str, Any]]
'''JSON-RPC params type'''  # for sphinx autodoc

JsonRpcRequestId = Union[str, int]
'''JSON-RPC identifier type'''  # for sphinx autodoc

Json = Union[List[Any], Tuple[Any, ...], Set[Any], Dict[str, Any], None]
'''JSON type'''  # for sphinx autodoc

MethodType = Callable[..., Any]
'''Method type'''  # for sphinx autodoc

Func = TypeVar('Func', bound=MethodType)
'''Function type'''  # for sphinx autodoc



import asyncio
import logging
import uuid
from typing import Any, Dict, Optional, cast

import aio_pika
from aio_pika.abc import AbstractIncomingMessage
from yarl import URL

import pjrpc
from pjrpc.client import AbstractAsyncClient

logger = logging.getLogger(__package__)


class Client(AbstractAsyncClient):
    """
    `aio_pika <http://kombu.readthedocs.org/>`_ based JSON-RPC client.

    :param broker_url: broker connection url
    :param conn_args: broker connection arguments.
    :param queue_name: queue name to publish requests to
    :param exchange_name: exchange to publish requests to. If ``None`` default exchange is used
    :param exchange_args: exchange arguments
    :param routing_key: reply message routing key. If ``None`` queue name is used
    :param result_queue_name: result queue name. If ``None`` random exclusive queue is declared for each request
    :param conn_args: additional connection arguments
    :param kwargs: parameters to be passed to :py:class:`pjrpc.client.AbstractClient`
    """

    def __init__(
        self,
        broker_url: URL,
        queue_name: Optional[str] = None,
        conn_args: Optional[Dict[str, Any]] = None,
        exchange_name: Optional[str] = None,
        exchange_args: Optional[Dict[str, Any]] = None,
        routing_key: Optional[str] = None,
        result_queue_name: Optional[str] = None,
        result_queue_args: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        assert queue_name or routing_key, "queue_name or routing_key must be provided"

        super().__init__(**kwargs)
        self._connection = aio_pika.connection.Connection(broker_url, **(conn_args or {}))
        self._channel: Optional[aio_pika.abc.AbstractChannel] = None

        self._exchange_name = exchange_name
        self._exchange_args = exchange_args
        self._exchange: Optional[aio_pika.abc.AbstractExchange] = None

        self._routing_key = cast(str, routing_key or queue_name)
        self._result_queue_name = result_queue_name
        self._result_queue_args = result_queue_args
        self._result_queue: Optional[aio_pika.abc.AbstractQueue] = None
        self._consumer_tag: Optional[str] = None

        self._futures: Dict[str, asyncio.Future[str]] = {}

    async def connect(self) -> None:
        """
        Opens a connection to the broker.
        """

        await self._connection.connect()
        self._channel = channel = await self._connection.channel()

        if self._exchange_name:
            self._exchange = await channel.declare_exchange(self._exchange_name, **(self._exchange_args or {}))

        if self._result_queue_name:
            assert channel
            self._result_queue = await channel.declare_queue(
                self._result_queue_name, **(self._result_queue_args or {}),
            )
            self._consumer_tag = await self._result_queue.consume(self._on_result_message, no_ack=True)

    async def close(self) -> None:
        """
        Closes current broker connection.
        """

        if self._consumer_tag and self._result_queue:
            await self._result_queue.cancel(self._consumer_tag)
            self._consumer_tag = None

        if self._channel:
            await self._channel.close()
            self._channel = None
        if self._connection:
            await self._connection.close()
            self._connection = None

        for future in self._futures.values():
            if future.done():
                continue

            future.set_exception(asyncio.CancelledError)

    async def _on_result_message(self, message: AbstractIncomingMessage) -> None:
        correlation_id = message.correlation_id
        assert correlation_id
        future = self._futures.pop(correlation_id, None)

        if future is None:
            logger.warning("unexpected or outdated message received: %r", message)
            return

        if message.content_type not in pjrpc.common.RESPONSE_CONTENT_TYPES:
            future.set_exception(
                pjrpc.exc.DeserializationError(f"unexpected response content type: {message.content_type}"),
            )
        else:
            future.set_result(message.body.decode(message.content_encoding or 'utf8'))

    async def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        if is_notification:
            async with self._connection.channel() as channel:
                message = aio_pika.message.Message(
                    body=request_text.encode(),
                    content_encoding='utf8',
                    content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                    **kwargs,
                )
                exchange = self._exchange or channel.default_exchange
                await exchange.publish(message, routing_key=self._routing_key)
                return None

        request_id = str(uuid.uuid4())

        async with self._connection.channel() as channel:
            if not self._result_queue:
                result_queue = await channel.declare_queue(
                    request_id, exclusive=True, **(self._result_queue_args or {}),
                )
                await result_queue.consume(self._on_result_message, no_ack=True)
            else:
                result_queue = self._result_queue

            message = aio_pika.message.Message(
                body=request_text.encode(),
                correlation_id=request_id,
                reply_to=result_queue.name,
                content_encoding='utf8',
                content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                **kwargs,
            )

            future: asyncio.Future[str] = asyncio.Future()
            self._futures[request_id] = future

            try:
                exchange = self._exchange or channel.default_exchange
                await exchange.publish(message, routing_key=self._routing_key)
                return await future
            finally:
                self._futures.pop(request_id, None)



import inspect
from typing import Any, Dict, Iterable, List, Optional, Type

import pydantic as pd

from pjrpc.common import UNSET, MaybeSet
from pjrpc.common.exceptions import JsonRpcError
from pjrpc.common.typedefs import MethodType
from pjrpc.server.specs.extractors import BaseSchemaExtractor, Error, Schema


class PydanticSchemaExtractor(BaseSchemaExtractor):
    """
    Pydantic method specification extractor.
    """

    def __init__(self, ref_template: str = '#/components/schemas/{model}'):
        self._ref_template = ref_template

    def extract_params_schema(self, method: MethodType, exclude: Iterable[str] = ()) -> Dict[str, Schema]:
        exclude = set(exclude)
        signature = inspect.signature(method)

        field_definitions: Dict[str, Any] = {}
        for param in signature.parameters.values():
            if param.name in exclude:
                continue

            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        params_model = pd.create_model('RequestModel', **field_definitions)
        model_schema = params_model.schema(ref_template=self._ref_template)

        parameters_schema = {}
        for param_name, param_schema in model_schema['properties'].items():
            required = param_name in model_schema.get('required', [])

            parameters_schema[param_name] = Schema(
                schema=param_schema,
                summary=param_schema.get('title', UNSET),
                description=param_schema.get('description', UNSET),
                deprecated=param_schema.get('deprecated', UNSET),
                required=required,
                definitions=model_schema.get('definitions'),
            )

        return parameters_schema

    def extract_result_schema(self, method: MethodType) -> Schema:
        result = inspect.signature(method)

        if result.return_annotation is inspect.Parameter.empty:
            return_annotation = Any
        elif result.return_annotation is None:
            return_annotation = Optional[None]
        else:
            return_annotation = result.return_annotation

        result_model = pd.create_model('ResultModel', result=(return_annotation, pd.fields.Undefined))
        model_schema = result_model.schema(ref_template=self._ref_template)

        result_schema = model_schema['properties']['result']
        required = 'result' in model_schema.get('required', [])
        if not required:
            result_schema['nullable'] = 'true'

        result_schema = Schema(
            schema=result_schema,
            summary=result_schema.get('title', UNSET),
            description=result_schema.get('description', UNSET),
            deprecated=result_schema.get('deprecated', UNSET),
            required=required,
            definitions=model_schema.get('definitions', UNSET),
        )

        return result_schema

    def extract_errors_schema(
        self,
        method: MethodType,
        errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> MaybeSet[List[Error]]:
        if errors:
            errors_schema = []
            for error in errors:
                field_definitions: Dict[str, Any] = {}
                for field_name, annotation in self._get_annotations(error).items():
                    if field_name.startswith('_'):
                        continue

                    field_definitions[field_name] = (annotation, getattr(error, field_name, ...))

                result_model = pd.create_model(error.message, **field_definitions)
                model_schema = result_model.schema(ref_template=self._ref_template)

                data_schema = model_schema['properties'].get('data', UNSET)
                required = 'data' in model_schema.get('required', [])

                errors_schema.append(
                    Error(
                        code=error.code,
                        message=error.message,
                        data=data_schema,
                        data_required=required,
                        title=error.message,
                        description=inspect.cleandoc(error.__doc__) if error.__doc__ is not None else UNSET,
                        deprecated=model_schema.get('deprecated', UNSET),
                        definitions=model_schema.get('definitions'),
                    ),
                )
            return errors_schema

        else:
            return UNSET

    @staticmethod
    def _extract_field_schema(model_schema: Dict[str, Any], field_name: str) -> Dict[str, Any]:
        field_schema = model_schema['properties'][field_name]
        if '$ref' in field_schema:
            field_schema = model_schema['definitions'][field_schema['$ref']]

        return field_schema

    @staticmethod
    def _get_annotations(cls: Type[Any]) -> Dict[str, Any]:
        annotations: Dict[str, Any] = {}
        for patent in cls.mro():
            annotations.update(**getattr(patent, '__annotations__', {}))

        return annotations



import asyncio
import functools as ft
import itertools as it
import json
import logging
from typing import Any, Awaitable, Callable, Dict, Generator, ItemsView, Iterable, Iterator, KeysView, List, Optional
from typing import Type, Union, ValuesView, cast

import pjrpc
from pjrpc.common import UNSET, AbstractResponse, BatchRequest, BatchResponse, MaybeSet, Request, Response, UnsetType
from pjrpc.common import v20
from pjrpc.common.typedefs import JsonRpcParams, MethodType
from pjrpc.server import utils
from pjrpc.server.typedefs import AsyncErrorHandlerType, AsyncMiddlewareType, ErrorHandlerType, MiddlewareType

from . import validators

logger = logging.getLogger(__package__)

default_validator = validators.base.BaseValidator()


class Method:
    """
    JSON-RPC method wrapper. Stores method itself and some metainformation.

    :param method: method
    :param name: method name
    :param context: context name
    """

    def __init__(self, method: MethodType, name: Optional[str] = None, context: Optional[str] = None):
        self.method = method
        self.name = name or method.__name__
        self.context = context

        meta = utils.set_meta(method, method_name=self.name, context_name=context)

        self.validator, self.validator_args = meta.get('validator', default_validator), meta.get('validator_args', {})

    def bind(self, params: Optional['JsonRpcParams'], context: Optional[Any] = None) -> MethodType:
        method_params = self.validator.validate_method(
            self.method, params, exclude=(self.context,) if self.context else (), **self.validator_args,
        )

        if self.context is not None:
            method_params[self.context] = context

        return ft.partial(self.method, **method_params)

    def copy(self, **kwargs: Any) -> 'Method':
        cls_kwargs = dict(name=self.name, context=self.context)
        cls_kwargs.update(kwargs)

        return Method(method=self.method, **cls_kwargs)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Method):
            return NotImplemented

        return (self.method, self.name, self.context) == (other.method, other.name, other.context)


class ViewMethod(Method):
    """
    View method.

    :param view_cls: view class
    :param name: view class method name
    :param context: context name
    """

    def __init__(
        self,
        view_cls: Type['ViewMixin'],
        method_name: str,
        name: Optional[str] = None,
        context: Optional[Any] = None,
    ):
        super().__init__(getattr(view_cls, method_name), name or method_name, context)

        self.view_cls = view_cls
        self.method_name = method_name

    def bind(self, params: Optional['JsonRpcParams'], context: Optional[Any] = None) -> MethodType:
        view = self.view_cls(context) if self.context else self.view_cls()
        method = getattr(view, self.method_name)

        method_params = self.validator.validate_method(method, params, **self.validator_args)

        return ft.partial(method, **method_params)

    def copy(self, **kwargs: Any) -> 'ViewMethod':
        cls_kwargs = dict(name=self.name, context=self.context)
        cls_kwargs.update(kwargs)

        return ViewMethod(view_cls=self.view_cls, method_name=self.method_name, **cls_kwargs)


class ViewMixin:
    """
    Simple class based method handler mixin. Exposes all public methods.
    """

    def __init__(self, context: Optional[Any] = None):
        pass

    @classmethod
    def __methods__(cls) -> Generator[Callable[..., Any], None, None]:
        for attr_name in filter(lambda name: not name.startswith('_'), dir(cls)):
            attr = getattr(cls, attr_name)
            if callable(attr):
                yield attr


class MethodRegistry:
    """
    Method registry.

    :param prefix: method name prefix to be used for naming containing methods
    """

    def __init__(self, prefix: Optional[str] = None):
        self._prefix = prefix
        self._registry: Dict[str, Method] = {}

    def __iter__(self) -> Iterator[str]:
        """
        Returns registry method iterator.
        """

        return iter(self._registry)

    def __getitem__(self, item: str) -> Method:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method
        :raises: KeyError
        """

        return self._registry[item]

    def items(self) -> ItemsView[str, Method]:
        return self._registry.items()

    def keys(self) -> KeysView[str]:
        return self._registry.keys()

    def values(self) -> ValuesView[Method]:
        return self._registry.values()

    def get(self, item: str) -> Optional[Method]:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method or `None`
        """

        return self._registry.get(item)

    def add(
        self, maybe_method: Optional[MethodType] = None, name: Optional[str] = None, context: Optional[Any] = None,
    ) -> MethodType:
        """
        Decorator adding decorated method to the registry.

        :param maybe_method: method or `None`
        :param name: method name to be used instead of `__name__` attribute
        :param context: parameter name to be used as an application context
        :returns: decorated method or decorator
        """

        def decorator(method: MethodType) -> MethodType:
            full_name = '.'.join(filter(None, (self._prefix, name or method.__name__)))
            self.add_methods(Method(method, full_name, context))

            return method

        if maybe_method is None:
            return decorator
        else:
            return decorator(maybe_method)

    def add_methods(self, *methods: Union[Callable[..., Any], Method]) -> None:
        """
        Adds methods to the registry.

        :param methods: methods to be added. Each one can be an instance of :py:class:`pjrpc.server.Method`
                        or plain method
        """

        for method in methods:
            if isinstance(method, Method):
                self._add_method(method)
            else:
                self.add(method)

    def view(
        self, maybe_view: Optional[Type[ViewMixin]] = None, context: Optional[Any] = None, prefix: Optional[str] = None,
    ) -> Union[ViewMixin, Callable[..., Any]]:
        """
        Methods view decorator.

        :param maybe_view: view class instance or `None`
        :param context: application context name
        :param prefix: view methods prefix
        :return: decorator or decorated view
        """

        def decorator(view: Type[ViewMixin]) -> Type[ViewMixin]:
            for method in view.__methods__():
                full_name = '.'.join(filter(None, (self._prefix, prefix, method.__name__)))
                self._add_method(ViewMethod(view, method.__name__, full_name, context))

            return view

        # maybe_view's type depends on the usage of the decorator.  It's a View
        # if it's used as `@view` but ``None`` if used as `@view()`.
        if maybe_view is None:
            return decorator
        else:
            return decorator(maybe_view)

    def merge(self, other: 'MethodRegistry') -> None:
        """
        Merges two registries.

        :param other: registry to be merged in the current one
        """

        for name, method in other.items():
            if self._prefix:
                name = f'{self._prefix}.{name}'

            self._add_method(method.copy(name=name))

    def _add_method(self, method: Method) -> None:
        if method.name in self._registry:
            logger.warning(f"method '{method.name}' already registered")

        self._registry[method.name] = method


class JSONEncoder(pjrpc.JSONEncoder):
    """
    Server JSON encoder. All custom server encoders should be inherited from it.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, validators.base.ValidationError):
            return [err for err in o.args]

        return super().default(o)


class BaseDispatcher:
    """
    Method dispatcher.

    :param request_class: JSON-RPC request class
    :param response_class: JSON-RPC response class
    :param batch_request: JSON-RPC batch request class
    :param batch_response: JSON-RPC batch response class
    :param json_loader: request json loader
    :param json_dumper: response json dumper
    :param json_encoder: response json encoder
    :param json_decoder: request json decoder
    :param middlewares: request middlewares
    :param error_handlers: request error handlers
    """

    def __init__(
        self,
        *,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request: Type[BatchRequest] = v20.BatchRequest,
        batch_response: Type[BatchResponse] = v20.BatchResponse,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: Type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[Type[json.JSONDecoder]] = None,
        middlewares: Iterable[Callable[..., Any]] = (),
        error_handlers: Dict[Union[None, int, Exception], List[Callable[..., Any]]] = {},
    ):
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._request_class = request_class
        self._response_class = response_class
        self._batch_request = batch_request
        self._batch_response = batch_response
        self._middlewares = list(middlewares)
        self._error_handlers = error_handlers

        self._registry = MethodRegistry()

    @property
    def registry(self) -> MethodRegistry:
        return self._registry

    def add(self, method: Callable[..., Any], name: Optional[str] = None, context: Optional[Any] = None) -> None:
        """
        Adds method to the registry.

        :param method: method
        :param name: method name
        :param context: application context name
        """

        self._registry.add(method, name, context)

    def add_methods(self, *methods: Union[MethodRegistry, Method, Callable[..., Any]]) -> None:
        """
        Adds methods to the registry.

        :param methods: method list. Each method may be an instance of :py:class:`pjrpc.server.MethodRegistry`,
                        :py:class:`pjrpc.server.Method` or plain function
        """

        for method in methods:
            if isinstance(method, MethodRegistry):
                self._registry.merge(method)
            elif isinstance(method, Method):
                self._registry.add_methods(method)
            else:
                self._registry.add(method)

    def view(self, view: Type[ViewMixin]) -> None:
        """
        Adds class based view to the registry.

        :param view: view to be added
        """

        self._registry.view(view)


class Dispatcher(BaseDispatcher):
    """
    Synchronous method dispatcher.
    """

    def __init__(
        self,
        *,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request: Type[BatchRequest] = v20.BatchRequest,
        batch_response: Type[BatchResponse] = v20.BatchResponse,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: Type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[Type[json.JSONDecoder]] = None,
        middlewares: Iterable['MiddlewareType'] = (),
        error_handlers: Dict[Union[None, int, Exception], List['ErrorHandlerType']] = {},
    ):
        super().__init__(
            request_class=request_class,
            response_class=response_class,
            batch_request=batch_request,
            batch_response=batch_response,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
            error_handlers=error_handlers,
        )

    def dispatch(self, request_text: str, context: Optional[Any] = None) -> Optional[str]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.getChild('request').debug("request received: %s", request_text)

        response: MaybeSet[AbstractResponse]
        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            request: Union[Request, BatchRequest]
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, BatchRequest):
                response = self._batch_response(
                    *(
                        resp for resp in (self._handle_request(request, context) for request in request)
                        if not isinstance(resp, UnsetType)
                    ),
                )
            else:
                response = self._handle_request(request, context)

        if not isinstance(response, UnsetType):
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.getChild('response').debug("response sent: %s", response_text)

            return response_text

        return None

    def _handle_request(self, request: Request, context: Optional[Any]) -> MaybeSet[Response]:
        try:
            HandlerType = Callable[[Request, Optional[Any]], MaybeSet[Response]]
            handler: HandlerType = self._handle_rpc_request

            for middleware in reversed(self._middlewares):
                handler = cast(HandlerType, ft.partial(middleware, handler=handler))

            return handler(request, context)

        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        for error_handler in it.chain(self._error_handlers.get(None, []), self._error_handlers.get(error.code, [])):
            error = error_handler(request, context, error)

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    def _handle_rpc_request(self, request: Request, context: Optional[Any]) -> MaybeSet[Response]:
        result = self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    def _handle_rpc_method(
        self,
        method_name: str,
        params: Optional[JsonRpcParams],
        context: Optional[Any],
    ) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            bound_method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            return bound_method()

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e


class AsyncDispatcher(BaseDispatcher):
    """
    Asynchronous method dispatcher.
    """

    def __init__(
        self,
        *,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request: Type[BatchRequest] = v20.BatchRequest,
        batch_response: Type[BatchResponse] = v20.BatchResponse,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: Type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[Type[json.JSONDecoder]] = None,
        middlewares: Iterable['AsyncMiddlewareType'] = (),
        error_handlers: Dict[Union[None, int, Exception], List['AsyncErrorHandlerType']] = {},
    ):
        super().__init__(
            request_class=request_class,
            response_class=response_class,
            batch_request=batch_request,
            batch_response=batch_response,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
            error_handlers=error_handlers,
        )

    async def dispatch(self, request_text: str, context: Optional[Any] = None) -> Optional[str]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.getChild('request').debug("request received: %s", request_text)

        response: MaybeSet[AbstractResponse]
        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            request: Union[Request, BatchRequest]
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, BatchRequest):
                response = self._batch_response(
                    *filter(
                        lambda resp: resp is not UNSET, await asyncio.gather(
                            *(self._handle_request(request, context) for request in request),
                        ),
                    ),
                )

            else:
                response = await self._handle_request(request, context)

        if not isinstance(response, UnsetType):
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.getChild('response').debug("response sent: %s", response_text)

            return response_text

        return None

    async def _handle_request(self, request: Request, context: Optional[Any]) -> MaybeSet[Response]:
        try:
            HandlerType = Callable[[Request, Optional[Any]], Awaitable[MaybeSet[Response]]]
            handler: HandlerType = self._handle_rpc_request

            for middleware in reversed(self._middlewares):
                handler = cast(HandlerType, ft.partial(middleware, handler=handler))

            return await handler(request, context)

        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        for error_handler in it.chain(self._error_handlers.get(None, []), self._error_handlers.get(error.code, [])):
            error = await error_handler(request, context, error)

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    async def _handle_rpc_request(self, request: Request, context: Optional[Any]) -> MaybeSet[Response]:
        result = await self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    async def _handle_rpc_method(
        self, method_name: str, params: Optional[JsonRpcParams], context: Optional[Any],
    ) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            bound_method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            result = bound_method()
            if asyncio.iscoroutine(result):
                result = await result

            return result

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e



"""
OpenRPC specification generator. See https://spec.open-rpc.org/.
"""
import enum

try:
    import dataclasses as dc
except ImportError:
    raise AssertionError("python 3.7 or later is required")

import itertools as it
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Type, Union

from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions
from pjrpc.common.typedefs import Func
from pjrpc.server import Method, utils

from . import Specification, extractors

Json = Union[str, int, float, dict, bool, list, tuple, set, None]  # type: ignore[type-arg]


@dc.dataclass(frozen=True)
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET
    email: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param url: a URL to the license used for the API
    """

    name: str
    url: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenRPC document
    :param description: a verbose description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    description: MaybeSet[str] = UNSET
    contact: MaybeSet[Contact] = UNSET
    license: MaybeSet[License] = UNSET
    termsOfService: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class Server:
    """
    Connectivity information of a target server.

    :param name: a name to be used as the canonical name for the server.
    :param url: a URL to the target host
    :param summary: a short summary of what the server is
    :param description: an optional string describing the host designated by the URL
    """

    name: str
    url: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: A verbose explanation of the target documentation
    :param description: The URL for the target documentation. Value MUST be in the format of a URL
    """

    url: str
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param summary: a short summary of the tag
    :param description: a verbose explanation for the tag
    :param externalDocs: additional external documentation for this tag
    """

    name: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET


@dc.dataclass(frozen=True)
class ExampleObject:
    """
    The ExampleObject object is an object the defines an example.

    :param value: embedded literal example
    :param name: canonical name of the example
    :param summary: short description for the example
    :param description: a verbose explanation of the example
    """

    value: Json
    name: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class MethodExample:
    """
    The example Pairing object consists of a set of example params and result.

    :param params: example parameters
    :param result: example result
    :param name: name for the example pairing
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    name: str
    params: List[ExampleObject]
    result: ExampleObject
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class ContentDescriptor:
    """
    Content Descriptors are objects that describe content.
    They are reusable ways of describing either parameters or result.

    :param name: name of the content that is being described
    :param schema: schema that describes the content. The Schema Objects MUST follow the specifications outline
                   in the JSON Schema Specification 7 (https://json-schema.org/draft-07/json-schema-release-notes.html)
    :param summary: a short summary of the content that is being described
    :param description: a verbose explanation of the content descriptor behavior
    :param required: determines if the content is a required field
    :param deprecated: specifies that the content is deprecated and SHOULD be transitioned out of usage
    """

    name: str
    schema: Dict[str, Any]
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    required: MaybeSet[bool] = UNSET
    deprecated: MaybeSet[bool] = UNSET


@dc.dataclass(frozen=True)
class Error:
    """
    Defines an application level error.

    :param code: a Number that indicates the error type that occurred
    :param message: a String providing a short description of the error
    :param data: a Primitive or Structured value that contains additional information about the error
    """

    code: int
    message: str
    data: MaybeSet[Dict[str, Any]] = UNSET


class ParamStructure(str, enum.Enum):
    """
    The expected format of the parameters.
    """

    BY_NAME = 'by-name'
    BY_POSITION = 'by-position'
    EITHER = 'either'


@dc.dataclass(frozen=True)
class MethodInfo:
    """
    Describes the interface for the given method name.

    :param name: the canonical name for the method
    :param params: a list of parameters that are applicable for this method
    :param result: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage examples
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param paramStructure: the expected format of the parameters
    :param externalDocs: additional external documentation for this method
    :param servers: an alternative servers array to service this method
    """

    name: str
    params: List[Union[ContentDescriptor, Dict[str, Any]]]
    result: Union[ContentDescriptor, Dict[str, Any]]
    errors: MaybeSet[List[Error]] = UNSET
    paramStructure: MaybeSet[ParamStructure] = UNSET
    examples: MaybeSet[List[MethodExample]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    tags: MaybeSet[List[Tag]] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    servers: MaybeSet[List[Server]] = UNSET


@dc.dataclass(frozen=True)
class Components:
    """
    Set of reusable objects for different aspects of the OpenRPC.

    :param schemas: reusable Schema Objects
    """

    schemas: Dict[str, Any] = dc.field(default_factory=dict)


def annotate(
    params_schema: MaybeSet[List[ContentDescriptor]] = UNSET,
    result_schema: MaybeSet[ContentDescriptor] = UNSET,
    errors: MaybeSet[List[Union[Error, Type[exceptions.JsonRpcError]]]] = UNSET,
    examples: MaybeSet[List[MethodExample]] = UNSET,
    summary: MaybeSet[str] = UNSET,
    description: MaybeSet[str] = UNSET,
    tags: MaybeSet[List[Union[Tag, str]]] = UNSET,
    deprecated: MaybeSet[bool] = UNSET,
) -> Callable[[Func], Func]:
    """
    Adds JSON-RPC method to the API specification.

    :param params_schema: a list of parameters that are applicable for this method
    :param result_schema: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage example
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    """

    def decorator(method: Func) -> Func:
        utils.set_meta(
            method,
            openrpc_spec=dict(
                params_schema=params_schema,
                result_schema=result_schema,
                errors=[
                    error if isinstance(error, Error) else Error(code=error.code, message=error.message)
                    for error in errors
                ] if errors else UNSET,
                examples=examples,
                tags=[
                    Tag(name=tag) if isinstance(tag, str) else tag
                    for tag in tags
                ] if tags else UNSET,
                summary=summary,
                description=description,
                deprecated=deprecated,
            ),
        )

        return method

    return decorator


@dc.dataclass(init=False)
class OpenRPC(Specification):
    """
    OpenRPC Specification.

    :param info: specification information
    :param path: specification url path
    :param servers: connectivity information
    :param external_docs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    :param schema_extractor: method specification extractor
    """

    info: Info
    components: Components
    methods: List[MethodInfo] = dc.field(default_factory=list)
    servers: MaybeSet[List[Server]] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    openrpc: str = '1.0.0'

    def __init__(
        self,
        info: Info,
        path: str = '/openrpc.json',
        servers: MaybeSet[List[Server]] = UNSET,
        external_docs: MaybeSet[ExternalDocumentation] = UNSET,
        openrpc: str = '1.0.0',
        schema_extractor: Optional[extractors.BaseSchemaExtractor] = None,
    ):
        super().__init__(path)

        self.info = info
        self.servers = servers
        self.externalDocs = external_docs
        self.openrpc = openrpc
        self.methods = []
        self.components = Components()

        self._schema_extractor = schema_extractor or extractors.BaseSchemaExtractor()

    def schema(
        self,
        path: str,
        methods: Iterable[Method] = (),
        methods_map: Mapping[str, Iterable[Method]] = {},
    ) -> Dict[str, Any]:
        for method in it.chain(methods, methods_map.get('', [])):
            method_name = method.name

            method_meta = utils.get_meta(method.method)
            annotated_spec = method_meta.get('openrpc_spec', {})

            params_schema = self._schema_extractor.extract_params_schema(
                method.method,
                exclude=[method.context] if method.context else [],
            )
            result_schema = self._schema_extractor.extract_result_schema(method.method)
            extracted_spec: Dict[str, Any] = dict(
                params_schema=[
                    ContentDescriptor(
                        name=name,
                        schema=schema.schema,
                        summary=schema.summary,
                        description=schema.description,
                        required=schema.required,
                        deprecated=schema.deprecated,
                    ) for name, schema in params_schema.items()
                ],
                result_schema=ContentDescriptor(
                    name='result',
                    schema=result_schema.schema,
                    summary=result_schema.summary,
                    description=result_schema.description,
                    required=result_schema.required,
                    deprecated=result_schema.deprecated,
                ),
                errors=self._schema_extractor.extract_errors_schema(method.method),
                deprecated=self._schema_extractor.extract_deprecation_status(method.method),
                description=self._schema_extractor.extract_description(method.method),
                summary=self._schema_extractor.extract_summary(method.method),
                tags=self._schema_extractor.extract_tags(method.method),
                examples=[
                    MethodExample(
                        name=example.summary or f'Example#{i}',
                        params=[
                            ExampleObject(
                                value=param_value,
                                name=param_name,
                            )
                            for param_name, param_value in example.params.items()
                        ],
                        result=ExampleObject(
                            name='result',
                            value=example.result,
                        ),
                        summary=example.summary,
                        description=example.description,
                    )
                    for i, example in enumerate(self._schema_extractor.extract_examples(method.method) or [])
                ],
            )
            method_spec: Dict[str, Any] = extracted_spec.copy()
            method_spec.update((k, v) for k, v in annotated_spec.items() if v is not UNSET)

            self.methods.append(
                MethodInfo(
                    name=method_name,
                    params=method_spec['params_schema'],
                    result=method_spec['result_schema'],
                    errors=method_spec['errors'],
                    examples=method_spec['examples'],
                    summary=method_spec['summary'],
                    description=method_spec['description'],
                    tags=method_spec['tags'],
                    deprecated=method_spec['deprecated'],
                ),
            )

            for param_schema in params_schema.values():
                if not isinstance(param_schema.definitions, UnsetType):
                    self.components.schemas.update(param_schema.definitions)

            if not isinstance(result_schema.definitions, UnsetType):
                self.components.schemas.update(result_schema.definitions)

        return dc.asdict(
            self,
            dict_factory=lambda iterable: dict(
                filter(lambda item: not isinstance(item[1], UnsetType), iterable),
            ),
        )



import dataclasses as dc
import inspect
import itertools as it
from typing import Any, Dict, Iterable, List, Optional, Type

from pjrpc.common import UNSET, MaybeSet, UnsetType
from pjrpc.common.exceptions import JsonRpcError
from pjrpc.common.typedefs import MethodType


@dc.dataclass(frozen=True)
class Schema:
    """
    Method parameter/result schema.
    """

    schema: Dict[str, Any]
    required: bool = True
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    definitions: MaybeSet[Dict[str, Any]] = UNSET


@dc.dataclass(frozen=True)
class Example:
    """
    Method usage example.
    """

    params: Dict[str, Any]
    result: Any
    version: str = '2.0'
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class ErrorExample:
    """
    Method error example.
    """

    code: int
    message: str
    data: MaybeSet[Optional[Any]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class Tag:
    """
    A list of method tags.
    """

    name: str
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[str] = UNSET


@dc.dataclass(frozen=True)
class Error:
    """
    Defines an application level error.
    """

    code: int
    message: str
    data: MaybeSet[Dict[str, Any]] = UNSET
    data_required: MaybeSet[bool] = UNSET
    title: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    definitions: MaybeSet[Dict[str, Any]] = UNSET


class BaseSchemaExtractor:
    """
    Base method schema extractor.
    """

    def extract_params_schema(self, method: MethodType, exclude: Iterable[str] = ()) -> Dict[str, Schema]:
        """
        Extracts method parameters schema.
        """

        return {}

    def extract_result_schema(self, method: MethodType) -> Schema:
        """
        Extracts method result schema.
        """

        return Schema(schema={})

    def extract_description(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method description.
        """

        description: MaybeSet[str]
        if method.__doc__:
            doc = inspect.cleandoc(method.__doc__)
            description = '\n'.join(it.takewhile(lambda line: line, doc.split('\n')))
        else:
            description = UNSET

        return description

    def extract_summary(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method summary.
        """

        description = self.extract_description(method)

        summary: MaybeSet[str]
        if not isinstance(description, UnsetType):
            summary = description.split('.')[0]
        else:
            summary = UNSET

        return summary

    def extract_errors_schema(
        self,
        method: MethodType,
        errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> MaybeSet[List[Error]]:
        """
        Extracts method errors schema.
        """

        return UNSET

    def extract_tags(self, method: MethodType) -> MaybeSet[List[Tag]]:
        """
        Extracts method tags.
        """

        return UNSET

    def extract_examples(self, method: MethodType) -> MaybeSet[List[Example]]:
        """
        Extracts method usage examples.
        """

        return UNSET

    def extract_error_examples(
        self,
        method: MethodType,
        errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> MaybeSet[List[ErrorExample]]:
        """
        Extracts method error examples.
        """

        return [
            ErrorExample(code=error.code, message=error.message, summary=error.message)
            for error in errors
        ] if errors else UNSET

    def extract_deprecation_status(self, method: MethodType) -> MaybeSet[bool]:
        """
        Extracts method deprecation status.
        """

        return UNSET



"""
Client and server common functions, types and classes that implements JSON-RPC protocol itself
and agnostic to any transport protocol layer (http, socket, amqp) and server-side implementation.
"""

from . import generators, typedefs
from .common import UNSET, JSONEncoder, MaybeSet, UnsetType
from .v20 import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, Request, Response

DEFAULT_CONTENT_TYPE = 'application/json'
'''default JSON-RPC client/server content type'''  # for sphinx autodoc

REQUEST_CONTENT_TYPES = ('application/json', 'application/json-rpc', 'application/jsonrequest')
'''allowed JSON-RPC server requests content types'''  # for sphinx autodoc

RESPONSE_CONTENT_TYPES = ('application/json', 'application/json-rpc')
'''allowed JSON-RPC client responses content types'''  # for sphinx autodoc


def set_default_content_type(content_type: str) -> None:
    """
    Sets default json-rpc client request / json-rpc server response content type.
    """

    global DEFAULT_CONTENT_TYPE

    DEFAULT_CONTENT_TYPE = content_type


__all__ = [
    'AbstractRequest',
    'AbstractResponse',
    'Request',
    'Response',
    'BatchRequest',
    'BatchResponse',
    'UNSET',
    'UnsetType',
    'MaybeSet',
    'JSONEncoder',
    'DEFAULT_CONTENT_TYPE',
    'REQUEST_CONTENT_TYPES',
    'RESPONSE_CONTENT_TYPES',
    'generators',
    'typedefs',
    'set_default_content_type',
]



"""
`pytest <https://docs.pytest.org/en/latest/>`_ client library integration.
Implements some utilities for mocking out ``pjrpc`` library clients.
"""

import asyncio
import collections
import functools as ft
import json
import unittest.mock
from types import ModuleType
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

import pytest

import pjrpc
from pjrpc import Response
from pjrpc.common import UNSET, MaybeSet
from pjrpc.common.typedefs import JsonRpcParams, JsonRpcRequestId

CallType = Dict[Tuple[str, str], unittest.mock.Mock]
MatchType = Dict[Tuple[str, str], List['Match']]


class Match:
    """
    Match object. Incorporates request matching information.
    """

    def __init__(
        self,
        endpoint: str,
        version: str,
        method_name: str,
        once: bool,
        callback: Optional[Callable[..., Any]],
        **response_data: Any,
    ):
        self.endpoint = endpoint
        self.version = version
        self.method_name = method_name
        self.once = once
        self.callback = callback
        self.response_data = response_data


class PjRpcMocker:
    """
    Synchronous JSON-RPC client mocker.

    :param target: method to be mocked
    :param mocker: mocking package
    :param passthrough: pass not mocked requests to the original method
    """

    def __init__(self, target: str, mocker: ModuleType = unittest.mock, passthrough: bool = False):
        self._target = target
        self._mocker = mocker
        self._patcher: Optional[Any] = None
        self._async_resp = False
        self._passthrough = passthrough

        self._matches: Dict[str, MatchType] = collections.defaultdict(lambda: collections.defaultdict(list))
        self._calls: Dict[str, CallType] = collections.defaultdict(dict)

    @property
    def calls(self) -> Dict[str, CallType]:
        """
        Dictionary of JSON-PRC method calls.
        """

        return self._calls

    def add(
        self,
        endpoint: str,
        method_name: str,
        result: MaybeSet[Any] = UNSET,
        error: MaybeSet[Any] = UNSET,
        id: Optional[JsonRpcRequestId] = None,
        version: str = '2.0',
        once: bool = False,
        callback: Optional[Callable[..., Any]] = None,
    ) -> None:
        """
        Appends response patch. If the same method patch already exists they will be used in a round-robin way.

        :param endpoint: request endpoint
        :param method_name: method name
        :param result: patched result
        :param error: patched error
        :param id: patched request id
        :param version: patched request version
        :param once: if ``True`` the patch will be deleted after the first call
        :param callback: patched request callback
        """

        match = Match(endpoint, version, method_name, once, id=id, result=result, error=error, callback=callback)
        self._matches[endpoint][(version, method_name)].append(match)

    def replace(
        self,
        endpoint: str,
        method_name: str,
        result: MaybeSet[Any] = UNSET,
        error: MaybeSet[Any] = UNSET,
        id: Optional[JsonRpcRequestId] = None,
        version: str = '2.0',
        once: bool = False,
        callback: Optional[Callable[..., Any]] = None,
        idx: int = 0,
    ) -> None:
        """
        Replaces a previously added response patch by a new one.

        :param endpoint: request endpoint
        :param method_name: method name
        :param result: patched result
        :param error: patched error
        :param id: patched request id
        :param version: patched request version
        :param once: if ``True`` the patch will be deleted after the first call
        :param callback: patched request callback
        :param idx: patch index (if there are more than one)
        """

        match = Match(endpoint, version, method_name, once, id=id, result=result, error=error, callback=callback)
        self._matches[endpoint][(version, method_name)][idx] = match

    def remove(
        self,
        endpoint: str,
        method_name: Optional[str] = None,
        version: str = '2.0',
    ) -> Union[MatchType, List[Match]]:
        """
        Removes a previously added response patch.

        :param endpoint: request endpoint
        :param method_name: method name
        :param version: JSON-RPC request version

        :returns: removed response patch
        """

        result: Union[MatchType, List[Match]]
        if method_name is None:
            result = self._matches.pop(endpoint)
        else:
            result = self._matches[endpoint].pop((version, method_name))

        self._cleanup_matches(endpoint, version, method_name)

        return result

    def reset(self) -> None:
        """
        Removes all added matches and reset call statistics.
        """

        self._matches.clear()
        for calls in self._calls.values():
            for stub in calls.values():
                stub.reset_mock()

        self._calls.clear()

    def start(self) -> Any:
        """
        Activates a patcher.
        """

        patcher = self._mocker.patch(self._target)
        with patcher:
            if asyncio.iscoroutinefunction(patcher.temp_original):
                self._async_resp = True

        side_effect: Callable[..., Any]
        if self._async_resp:
            async def side_effect(*args: Any, **kwargs: Any) -> str:
                return await self._on_request(*args, **kwargs)
        else:
            def side_effect(*args: Any, **kwargs: Any) -> str:
                return self._on_request(*args, **kwargs)

        self._patcher = self._mocker.patch(self._target, side_effect=side_effect, autospec=True)

        assert self._patcher is not None
        return self._patcher.start()

    def stop(self) -> None:
        """
        Stop an active patcher.
        """

        assert self._patcher is not None, 'mocker is not started'

        self.reset()
        self._patcher.stop()

    def _cleanup_matches(self, endpoint: str, version: str = '2.0', method_name: Optional[str] = None) -> None:
        matches: Optional[Union[MatchType, List[Match]]]
        if method_name is not None:
            matches = self._matches[endpoint].get((version, method_name))
        else:
            matches = self._matches[endpoint]

        if not matches and method_name:
            self._matches[endpoint].pop((version, method_name), None)
        if not self._matches[endpoint]:
            self._matches.pop(endpoint)

    def _on_request(self, origin_self: Any, request_text: str, is_notification: bool = False, **kwargs: Any) -> Any:
        assert self._patcher is not None, 'mocker is not started'

        endpoint = origin_self._endpoint
        matches = self._matches.get(endpoint)
        if matches is None:
            if self._passthrough:
                return self._patcher.temp_original(origin_self, request_text, is_notification, **kwargs)
            else:
                raise ConnectionRefusedError()

        json_data = json.loads(request_text)

        response: Union[pjrpc.BatchResponse, pjrpc.Response]
        if isinstance(json_data, (list, tuple)):
            response = pjrpc.BatchResponse()
            for request in pjrpc.BatchRequest.from_json(json_data):
                response.append(
                    self._match_request(endpoint, request.version, request.method, request.params, request.id),
                )

        else:
            request = pjrpc.Request.from_json(json_data)
            response = self._match_request(endpoint, request.version, request.method, request.params, request.id)

        if self._async_resp:
            async def wrapper() -> str:
                return json.dumps(response.to_json())
            return wrapper()
        else:
            return json.dumps(response.to_json())

    def _match_request(
        self,
        endpoint: str,
        version: str,
        method_name: str,
        params: Optional[JsonRpcParams],
        id: Optional[JsonRpcRequestId],
    ) -> Response:
        matches = self._matches[endpoint].get((version, method_name))
        if matches is None:
            return pjrpc.Response(id=id, error=pjrpc.exc.MethodNotFoundError(data=method_name))

        match = matches.pop(0)
        if not match.once:
            matches.append(match)

        self._cleanup_matches(endpoint, version, method_name)

        stub = self.calls[endpoint].setdefault(
            (version, method_name),
            self._mocker.MagicMock(spec=lambda *args, **kwargs: None, name=f'{endpoint}:{version}:{method_name}'),
        )
        if isinstance(params, (list, tuple)):
            stub(*params)
        elif isinstance(params, dict):
            stub(**params)
        else:
            stub(params)

        if match.callback:
            if isinstance(params, (list, tuple)):
                result = match.callback(*params)
            elif isinstance(params, dict):
                result = match.callback(**params)
            else:
                result = match.callback(params)

            return pjrpc.Response(id=id, result=result)

        else:
            return pjrpc.Response(
                id=id or match.response_data['id'],
                result=match.response_data['result'],
                error=match.response_data['error'],
            )

    def __enter__(self) -> 'PjRpcMocker':
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop()
        self.reset()


# shortcuts
PjRpcRequestsMocker = ft.partial(PjRpcMocker, target='pjrpc.client.backend.requests.Client._request')
PjRpcAiohttpMocker = ft.partial(PjRpcMocker, target='pjrpc.client.backend.aiohttp.Client._request')


@pytest.fixture
def pjrpc_requests_mocker() -> Generator[PjRpcMocker, None, None]:
    """
    Requests client mocking fixture.
    """

    with PjRpcRequestsMocker() as mocker:
        yield mocker


@pytest.fixture
def pjrpc_aiohttp_mocker() -> Generator[PjRpcMocker, None, None]:
    """
    Aiohttp client mocking fixture.
    """

    with PjRpcAiohttpMocker() as mocker:
        yield mocker



"""
Extensible `JSON-RPC <https://www.jsonrpc.org>`_ client/server library.
"""

from pjrpc.__about__ import __author__, __description__, __email__, __license__, __title__, __url__, __version__
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, Request, Response
from pjrpc.common import exceptions, set_default_content_type, typedefs

# shortcuts
exc = exceptions

__all__ = [
    '__title__',
    '__description__',
    '__url__',
    '__version__',
    '__author__',
    '__email__',
    '__license__',

    'exceptions',
    'exc',
    'typedefs',
    'set_default_content_type',
    'AbstractResponse',
    'AbstractRequest',
    'Request',
    'Response',
    'BatchRequest',
    'Response',
    'Response',
    'BatchResponse',
    'JSONEncoder',
]
